// rewritten by Zheng Gengbin
// perform sunroutine translation and interface

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include <stdio.h>

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

//extern "C" int strcasecmp(...);

extern Shadow_def * get_defined_shadow(SgSymbol *sym);

extern void distribute_array(SgStatement *,SgSymbol *,SgExpression *,Distribute_def *,Processor_def*,SgStatement *,int);
extern void analyse_array_ref(SgStatement *,SgExpression *,SgStatement *,SgSymbol *);
extern int compare_distribute_pattern(SgSymbol *, SgSymbol *);
extern int compare_shadow(Shadow_def *, Shadow_def *);

void analyse_dummy_mapping(SgStatement *s);
void add_dummy_mapping(SgStatement *);

extern void change_indo_scalar_reference(SgStatement * , independent_reference *);

int do_process_proc(SgStatement *s)
{
    SgCallStmt *call = isSgCallStmt(s);
    ASSERT(call);
    SgSymbol *funcName = call->name();
    if (funcName->attributes() & EXTERNAL_BIT ||
	funcName->attributes() & EXTRINSIC_HPF_BIT )
      return 0;
    else
      return 1;
}

int do_process_func(SgExpression *exp)
{
    SgFunctionCallExp *func = isSgFunctionCallExp(exp);
    ASSERT(func);
    SgSymbol *funcName = func->funName();
    if (funcName->attributes() & EXTERNAL_BIT ||
	funcName->attributes() & EXTRINSIC_HPF_BIT )
      return 0;
    else
      return 1;
}

void add_secdad(SgStatement * s, SgArrayRefExp * arrayexp, int counter)
{	
  SgSymbol * dadsec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
  
  int dim =0;
  for (int i=0; i<arrayexp->numberOfSubscripts(); i++) 
    if (!arrayexp->isScalarOnDim(i))
      dim++;
  
  SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
  
  SgStatement *assign = make_new_array_section(s, dadsec, dim , dad);
  s->insertStmtBefore(*assign, *s->controlParent());

}

void delete_secdad(SgStatement * s, SgArrayRefExp * arrayexp, int counter)
{
  SgSymbol * dadsec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,counter);
  SgStatement * funcall = make_delete_array(s , dadsec);
  s->insertStmtAfter(*funcall, *s->controlParent());
}

void define_secdad(SgStatement * s, SgArrayRefExp *arrayexp, int counter)
{

  int dimcounter=0;

  for (int i=0; i<arrayexp->numberOfSubscripts(); i++) {
    SgSymbol * arrayname= arrayexp->symbol();
    
    if (!arrayexp->isScalarOnDim(i)) {

      SgSymbol * sddsymb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,counter);
      dimcounter++;

      SgTripletExp * triplet=(SgTripletExp *)(arrayexp->subscript(i));
      SgExpression *start = triplet->item(0);
      if ( start == NULL) start = get_lbound(arrayname, i);
      analyse_expression(s, start);

      SgExpression *end = triplet->item(1);
      if ( end == NULL) end = get_ubound(arrayname, i);
      analyse_expression(s, end);

      SgExpression *stride = triplet->item(2);
      if ( stride == NULL) stride = new SgValueExp(1);
      analyse_expression(s, stride);

      SgSymbol * dad =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());

      SgStatement *funcall = make_set_array_triplet(s, sddsymb,dimcounter,
						    start,end,stride,dad,i+1);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
    else {
      SgSymbol * dadsymb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
      SgExpression * scalar = arrayexp ->subscript(i);
      SgSymbol * dad =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
      int dim = i+1;
      SgStatement *funcall = make_set_array_scalar(s, dadsymb, scalar, dad, dim);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
  }
}


// we need the first address of the template section;
SgExpression *get_first_address(SgStatement *s,SgArrayRefExp * arrayexp)
{
	SgSymbol *arraysymb = arrayexp->symbol(); ASSERT(arraysymb);

        SgExpression * address = NULL;
	int hasScalarDim=0;
	
	if ( is_out_of_core(arraysymb) ) {
	  address=new SgValueExp(0);
	  return address;
	}

	for (int j=0; j<arrayexp->numberOfSubscripts(); j++) {
	  if (arrayexp->subscript(j)->variant()!=DDOT) {
	    hasScalarDim=1;
	    break;
	  };
	}
	
	if (!hasScalarDim) {
	  address=new SgValueExp(0);
	}
	else {
	  // scalar array reference
	  for (int dimI = 0; dimI<arrayexp->numberOfSubscripts(); dimI++) {

	    SgExpression * multi=NULL;

	    if (arrayexp->subscript(dimI)->variant()!=DDOT) {
	      {
	      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
	      SgValueExp * dim = new SgValueExp(dimI+1);
	      SgExpression *rngfun = make_range(s,dad,dim);
		
	      SgExpression *dimExp = arrayexp->subscript(dimI)->copyPtr();
	      SgExpression *locfun = make_loc(s,rngfun,dimExp);
	      multi=locfun;
	      }
	      {		// why stride?
	      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
	      SgValueExp * dim = new SgValueExp(dimI+1);
	      SgExpression *strfun = make_stride(s,dad,dim);
	      multi= & ((*multi)*(*strfun));
	      }
	      address = address + (*multi);
	    }  // end of if(DDOT)
	  }  // end of for
	}
	return address;
}

// change variable(scalar)
void change_variable_dummy(SgStatement *s, SgExpression *var)
{
	SgSymbol *symb = var->symbol();
        if (is_pointer(symb)) change_variable_by_stack(s, var);
}

// change array ref(scalar)
SgArrayRefExp *change_scalar_array_ref_dummy(SgStatement *s, SgArrayRefExp *arrayexp)
{
	SgSymbol *symb = arrayexp->symbol();
        SgArrayRefExp *var = new SgArrayRefExp(*symb);
	SgExpression *address = get_first_address(s,arrayexp);
	var->addSubscript(*address);
        if (unknown_address(symb))
	   change_array_by_stack(s, var);	
	return var;
}

void change_procstmt(SgStatement *s)
{
  SgCallStmt * call = (SgCallStmt *)s;
  SgCallStmt * newcall = new SgCallStmt(*call->name());

  for (int i=0; i<call->numberOfArgs(); i++) {

    SgExpression * argexp = call->arg(i);

    switch(argexp->variant()) {
    case VAR_REF: {
	// if it is a scalar variable , leave it untouched !   zheng.gb
      if ( argexp->type()->variant() != T_ARRAY ) {
      	SgSymbol * argname = argexp->symbol();
      	SgExpression * var = new SgVarRefExp (argname);
	  // change variable dummy if neccessary.
	change_variable_dummy(s, var);
      	newcall->addArg(*var);

	if (i==0 && (call->name()->attributes() & INTRINSIC_FUNC_BIT) )
	{
		SgSymbol *dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", call->arg(i)->symbol()->identifier());
      		SgExpression * dadexp = new SgVarRefExp (dadsymb);
      		newcall->addArg(*dadexp);
	}
	break;
      }
	// else continue next step !
    };

    case ARRAY_REF: {
      SgArrayRefExp * arrayexp = (SgArrayRefExp *) argexp;
      SgSymbol * argname = s->getSymbolInCurrentScope(argexp->symbol()->identifier());
      
	// the DAD to insert in call statement
      SgSymbol *dadsymb;
      if (argexp->variant()==VAR_REF) 
		dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", argname->identifier());
      else
		dadsymb=s->getSymbolInCurrentScope(mixstring, "sdd", NULL, i+1);

	// if is scalar array refernce , we may create stack to store it
      if (isSgArrayType(arrayexp->type()) == NULL)
      {
	  // so here must be array_ref
	ASSERT(isSgArrayRefExp(arrayexp));
	  // if this is the proc call turned from intrinsic function call
          // the first arg will tbe turned to dad anyway
        if ( s->numberOfAttributes(INTRIN_SUB) > 0 && i==0 ) ;
        else if (argexp->numberOfAttributes(DUMMY_ADDR) == 0)
        {
	  // remap array scalar reference if necessary.
	  SgSymbol *dadsymb=s->getSymbolInCurrentScope(mixstring,"sdd",NULL,i+1);
	  analyse_array_ref(s,arrayexp,s,dadsymb);
          newcall->addArg(*arrayexp);
	  // assign return value to array.
	  dec_exp_level();
	  SgSymbol *tddsymb=s->getSymbolInCurrentScope(mixstring,"tdd",exp_level);
	  SgStatement * remap = make_remap(s , dadsymb , tddsymb);
          s->insertStmtAfter(*remap, *s->controlParent());
	  inc_exp_level();
	  break;
        }
      }

      // this step is to get right memory address 
      SgArrayRefExp * var = NULL;
      if (argexp->numberOfAttributes(DUMMY_ADDR) != 0 )
	var = (SgArrayRefExp *)argexp->attributeValue(0,DUMMY_ADDR);
      else {
	// check same argument appear more than once, then we must allocate more in-core mem.
        if (is_out_of_core(argname))
	{
          for (int j=i+1; j<call->numberOfArgs(); j++) {
	     SgExpression * arg1 = call->arg(j);
             SgSymbol * argname1 = arg1->symbol();
	     if (argname1 == argname) {
if (!silentFlag)
printf("Same out-of-core argument: '%s'.\n",argname->identifier());
		if (isSgVarRefExp(argexp)) 
		{
		  // copy the dad as sdd , so that we can assign a different in-core mem.
		SgSymbol *newdadsymb=s->getSymbolInCurrentScope(mixstring, "sdd", NULL, i+1);
		SgStatement *copydad =  make_new_array_copy(s,newdadsymb,dadsymb);
		s->insertStmtBefore(*copydad,*s->controlParent());
		  // delete_array
		SgStatement *deletestmt=make_delete_array(s,newdadsymb);
		s->insertStmtAfter(*deletestmt,*s->controlParent());
		dadsymb = newdadsymb;   // change dad
		}
		
		// allocate memory as in-core size. reset dad's in-core mem
		// set 'var' as the  stack variable: real_mb()
		SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", call->name()->identifier(), i);
      		SgExpression *in_core_exp = get_in_core_size(s);
#ifdef MEM_ALLOC
    		SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", NULL, i);
		SgStatement *allocstmt =  make_stack_alloc_size(s,argexp->type(),base,in_core_exp,mHandle);
    		SgSymbol * rs= s->getSymbolInCurrentScope(stackNameOfType(argexp->type()),"mb", NULL);
#else
		SgStatement *allocstmt = make_alloc(s,type,base,dadsymb);
    		SgSymbol * rs= s->getSymbolInCurrentScope(mixstring, stackNameOfType(type), NULL);
#endif
		s->insertStmtBefore(*allocstmt, *s->controlParent());

		{	// reset array base
		SgExpression *mhExp = new SgVarRefExp(mHandle);
		SgStatement *resetstmt = make_reset_in_core_base(s,dadsymb,argexp->type(),base,mhExp);
		s->insertStmtBefore(*resetstmt,*s->controlParent());
		}
		{
		SgStatement *freestmt=make_free(s,argexp->type(),mHandle);
		s->insertStmtAfter(*freestmt,*s->controlParent());
		}
    		var = new SgArrayRefExp(*rs);
    		SgExpression * baseexp= new SgVarRefExp(base);
    		var->addSubscript(*baseexp);
	     }
          } // for
	}  // if

	if (!var)
	var = change_scalar_array_ref_dummy(s, arrayexp);
      }
      newcall->addArg(*var);

        // this step is to insert the dad as dummy
 	// EXTERNAL function needn't add dad as dummy.
      if (do_process_proc(call) == 1) 
      {
        if (argexp->numberOfAttributes(DUMMY_NAME) != 0 )
		dadsymb = (SgSymbol *)argexp->attributeValue(0,DUMMY_NAME);
        SgExpression * dadexp = new SgVarRefExp (dadsymb);
        newcall->addArg(*dadexp);
      }

      break;
    };
    
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case BOOL_VAL:
    case CHAR_VAL:
    case STRING_VAL:
    case COMPLEX_VAL: {
      // * do nothing, just copy the old one;
      newcall->addArg(*argexp);
      break;
    };

    default:
//      printf("In change_procstmt, unexpected expression on line %d, id: %d, variant: %d\n", s->lineNumber(), argexp->id(), argexp->variant());    by zheng.gb
      // must be an scalar expression
      if (isSgArrayType(argexp->type()) == NULL ) 
      {
printf("change_procstmt: expression encountered . \n");
	analyse_array_ref_in_exp(s,argexp,s);
      	newcall->addArg(*argexp);
      }
      else
      	newcall->addArg(*argexp);
      break;
    };

  }
  // ????
  s->insertStmtAfter(*newcall, *s->controlParent());
}

void prepare_actualdad_for_proc(SgStatement *s,SgExpression *var,int &arraycounter)
{
    arraycounter++;

    switch (var->variant()) {
    case VAR_REF:
      break;

    case ARRAY_REF:
      {
	SgArrayRefExp * arrayexp = (SgArrayRefExp *) var;
	add_secdad(s, arrayexp, arraycounter);
	delete_secdad(s, arrayexp, arraycounter);
	define_secdad(s, arrayexp, arraycounter);
      }
      break;

    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case BOOL_VAL:
    case CHAR_VAL:
    case STRING_VAL:
    case COMPLEX_VAL: {
      // * do nothing;
      break;
    };
        
    default:
//      printf("In trans_procstmt, unexpected expression on line %d, id: %d, variant: %d\n", s->lineNumber(), var->id(), var->variant());	zheng.gb
      break;

    };
}

void trans_procstmt(SgStatement * s) 
{
  if (!do_process_proc(s)) return;

  SgCallStmt * call = (SgCallStmt *)s;

  int arraycounter=0;

  for (int i=0; i<call->numberOfArgs(); i++) {
    //printf("The id is %d\n", call->arg(i)->id());
    SgExpression * var = call->arg(i);
    // create dads
    prepare_actualdad_for_proc(s,var,arraycounter);
  };

  if ( s->numberOfAttributes(Proccall_Ref) )
	add_dummy_mapping(s);
  change_procstmt(s);

  s->deleteStmt();
}

//--------------------;

void change_prochedr(SgStatement * s)
{
  SgProcHedrStmt *  prochedr = (SgProcHedrStmt *) s;
  int org_numberOfParameters = prochedr->numberOfParameters();

//  SgSymbol * parameters[10];
  SgSymbol **parameters = new SgSymbol *[org_numberOfParameters];

  int i;

  for (i=0; i<org_numberOfParameters; i++) 
    parameters[i]=prochedr->parameter(i);

  for (i=0; i<org_numberOfParameters; i++) 
  {
    SgSymbol * parameter=parameters[i];
	// scalar variable skip		zheng.gb 1997/9/20
    if (isSgArrayType(parameter->type()) == NULL ) continue;
    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", parameter->identifier());
    prochedr->insertParameterAfter(*parameter, *dad);
  }
  delete [] parameters;
}

// add by zheng.gb
void change_funccall(SgStatement *s,SgExpression *exp)
{
  SgFunctionCallExp *call = isSgFunctionCallExp(exp);

  SgFunctionCallExp * newcall = new SgFunctionCallExp(*call->funName());

  for (int i=0; i<call->numberOfArgs(); i++) {

    SgExpression * argexp = call->arg(i);

    switch(argexp->variant()) {
    case VAR_REF: {
	// if it is a scalar variable , leave it untouched !   zheng.gb
      if ( argexp->type()->variant() != T_ARRAY ) {
      	SgSymbol * argname = call->arg(i)->symbol();
      	SgExpression * var = new SgVarRefExp (argname);
	  // change variable dummy if neccessary.
	change_variable_dummy(s, var);
      	newcall->addArg(*var);

	if (i==0 && (call->funName()->attributes() & INTRINSIC_FUNC_BIT) )
	{
		SgSymbol *dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", call->arg(i)->symbol()->identifier());
      		SgExpression * dadexp = new SgVarRefExp (dadsymb);
      		newcall->addArg(*dadexp);
	}
	break;
      }
	// else continue next step !
    };

    case ARRAY_REF: {
      SgArrayRefExp * arrayexp = (SgArrayRefExp *) argexp;
      SgSymbol * argname = s->getSymbolInCurrentScope(argexp->symbol()->identifier());
      
	// if is scalar array refernce , we may create stack to store it
      if (isSgArrayType(arrayexp->type()) == NULL)
      {
	  // so here must be array_ref
	ASSERT(isSgArrayRefExp(arrayexp));
	  // if this is the proc call turned from intrinsic function call
          // the first arg will tbe turned to dad anyway
        if ( s->numberOfAttributes(INTRIN_SUB) > 0 && i==0 ) ;
        else if (argexp->numberOfAttributes(DUMMY_ADDR) == 0)
        {
	  // remap array scalar reference if necessary.
	SgSymbol *dadsymb=s->getSymbolInCurrentScope(mixstring,"sdd",NULL,i+1);
	analyse_array_ref(s,arrayexp,s,dadsymb);
        newcall->addArg(*arrayexp);
	  // assign return value to array.
//        SgAssignStmt *ass = new SgAssignStmt(*oldarray,*arrayexp->copyPtr());
//        s->insertStmtAfter(*ass, *s->controlParent());
	dec_exp_level();
	SgSymbol *tddsymb=s->getSymbolInCurrentScope(mixstring,"tdd",exp_level);
	SgStatement * remap = make_remap(s , dadsymb , tddsymb);
        s->insertStmtAfter(*remap, *s->controlParent());
	inc_exp_level();
	break;
        }
      }

      SgArrayRefExp * var;
      
      if (argexp->numberOfAttributes(DUMMY_ADDR) != 0 )
	var = (SgArrayRefExp *)argexp->attributeValue(0,DUMMY_ADDR);
      else 
	var = change_scalar_array_ref_dummy(s, arrayexp);
      newcall->addArg(*var);

 	// EXTERNAL function needn't add dad as dummy.
      if (do_process_func(exp) == 1) 
      {
      SgSymbol *dadsymb;
      if (argexp->numberOfAttributes(DUMMY_NAME) != 0 )
		dadsymb = (SgSymbol *)argexp->attributeValue(0,DUMMY_NAME);
      else if (argexp->variant()==VAR_REF) 
	dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", call->arg(i)->symbol()->identifier());
      else
	dadsymb=s->getSymbolInCurrentScope(mixstring, "sdd", NULL, i+1);
      SgExpression * dadexp = new SgVarRefExp (dadsymb);
      newcall->addArg(*dadexp);
      }

      break;
    };
    
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case BOOL_VAL:
    case CHAR_VAL:
    case STRING_VAL:
    case COMPLEX_VAL: {
      // * do nothing, just copy the old one;
      newcall->addArg(*argexp);
      break;
    };

    default:
//      printf("In change_procstmt, unexpected expression on line %d, id: %d, variant: %d\n", s->lineNumber(), argexp->id(), argexp->variant());    by zheng.gb
      // must be an scalar expression
      if (isSgArrayType(argexp->type()) == NULL ) 
      {
printf("change_procstmt: expression encountered . \n");
	analyse_array_ref_in_exp(s,argexp,s);
      	newcall->addArg(*argexp);
      }
      else
      	newcall->addArg(*argexp);
      break;
    };

  }

  exp->setLhs(newcall->lhs());
}

// add by zheng.gb
// translate function call statement , same as proc call statement
void trans_funccall(SgStatement * s,SgExpression *funccall) 
{
  SgFunctionCallExp * call = (SgFunctionCallExp *)funccall;

  int arraycounter=0;

  for (int i=0; i<call->numberOfArgs(); i++) {
    //printf("The id is %d\n", call->arg(i)->id());
    SgExpression * var = call->arg(i);
    // create dads
    prepare_actualdad_for_proc(s,var,arraycounter);
  };

  if ( s->numberOfAttributes(Proccall_Ref) )
	add_dummy_mapping(s);

  // change_procstmt(s);
  change_funccall(s,funccall);

}


void analyse_proc_stat(SgStatement *s)
{
    if (!do_process_proc(s)) return;

    SgCallStmt *call = (SgCallStmt *)s;

    SgFunctionCallExp *proc_call = (SgFunctionCallExp *)s->expr(0);
	
    if ( s->numberOfAttributes(INDEP_Def) != 0 ){ //proc_stat in indo
		call->name()->addAttribute(INDEP_Def, NULL, 0);
    }

    if (proc_call->rhs())    // rhs point to the functuion matched in interface
    {
	proc_reference *proc_ref = new proc_reference();
	SgFunctionSymb *matchfunc = (SgFunctionSymb *)proc_call->rhs()->symbol(); ASSERT(matchfunc);
	SgFunctionSymb *realfunc = (SgFunctionSymb *)proc_call->funName(); ASSERT(realfunc);
	proc_ref->setHeader(matchfunc, realfunc);
	// ASSERT((realfunc->attributes() & INTERFACE_BIT) == 0);

	int num_of_args = proc_call->numberOfArgs();
	for ( int i=0; i<num_of_args ; i++ ) {
		  // simplified !!
		proc_ref->add_pair(proc_call->arg(i),matchfunc->parameter(i));
	}
	s->addAttribute(Proccall_Ref , proc_ref , sizeof(proc_ref) );

	// change the proc symbol to matchfunc
	if (matchfunc != realfunc)
		proc_call->setSymbol(*matchfunc);
    }

    analyse_dummy_mapping(s);

    if (proc_call->funName()->attributes() & INTRINSIC_FUNC_BIT )
    {
	// wjp  do intrinsic function search !
	replace_intrinsic_function_name(proc_call,1);

		// need add DAD decalration for this variable(result).
	SgExpression * argexp = proc_call->arg(0);
	if ( argexp->variant() == VAR_REF )
	{
		SgSymbol * argname = argexp->symbol();
		if ( argname->type()->variant() != T_ARRAY )
		{
		  if (argname->numberOfAttributes(NeedDad_Def)==0) 
  			argname->addAttribute(NeedDad_Def, NULL, 0);
		}
	}
    }
}

// check if we can use pcrc_same_data_mapping 
int check_reduced_dim_collapsed(SgExpression *exp)
{
	int use = 1;
	SgArrayRefExp *arrayexp = isSgArrayRefExp(exp);
	if (arrayexp == NULL) return use;
	SgSymbol *arraysymb = arrayexp->symbol();
	 // check scalar dim , if this dimension is not collapsed , remap is forced.
	for (int i=0; i<arrayexp->numberOfSubscripts(); i++)
	{
		if (arrayexp->subscript(i)->variant() == DDOT) continue;
		if (array_dim_collapsed(arraysymb,i) != 1) use=0;
	}
	return use;
}

SgSymbol *get_real_dummy(SgFunctionSymb *header, SgSymbol *dummy)
{
  for (int i=0; i<header->numberOfParameters(); i++)
	if (strcasecmp(dummy->identifier(), header->parameter(i)->identifier()) == 0) 
		return header->parameter(i);
  return NULL;
}

// analyse proc call statement , decide the remap pattern of each dummy.
void analyse_dummy_mapping(SgStatement *s)
{
	SgStatement *scope = s->getScopeForDeclare();

	for (int i=0; i < s->numberOfAttributes(Proccall_Ref);i++)
	{
		proc_reference *proc_ref = (proc_reference *)s->attributeValue(i,Proccall_Ref);

		SgStatement *interface_hedr = proc_ref->getHeader()->function_hedr();

		for (int j=0; j< proc_ref->get_number() ; j ++)
		{
		   int dummy_need_mapping = 0;	// 0: no need to remap
						// 1: use remap, INOUT, OUT	
						//-1: use remap, IN
						// 2: use pcrc_same_data_mapping to decide in runtime
						//-2: use pcrc_same_data_mapping to decide in runtime IN

		   SgSymbol *dummy = proc_ref->get_dummy(j);
		   SgExpression *exp = proc_ref->get_actual(j);

		   //   1. havd align 2. have distribute 3. have nothing
		   if (dummy->numberOfAttributes(Align_Def) != 0 )
		   {
			printf("Align dummy , crazy !\n");
		   }
		   else if (dummy->numberOfAttributes(Inherit_Def) != 0)
		   {
			dummy_need_mapping = 0;
		   }
		   else if (dummy->numberOfAttributes(Dist_Def) != 0 )
		   {
			dummy_need_mapping = 1;

			  // descriptive mapping
  	   		Distribute_def *ddef = (Distribute_def *) dummy->attributeValue(0, Dist_Def);

			// compare processor group
			SgArrayRefExp *dummyProcs = ddef->getProcs();
			if ( exp->variant() == VAR_REF ) 
			{
			    SgSymbol *symb = exp->symbol();
			    int status = same_processor_group(symb, dummy);
			    if (status == 0)
				dummy_need_mapping = 1;
			    else if (status == -1)
				dummy_need_mapping = 2;
			}

			  // distribute *(block)
			if (ddef->is_descriptive()) dummy_need_mapping = 0;

			if (opt_strategy & OTHER_OPTIMIZE)
			{
			    if (exp->variant() == VAR_REF) {
			      if (compare_distribute_pattern(exp->symbol(),dummy ) == 1)
			      {
if (!silentFlag)
printf("Actual arg '%s' and dummy '%s' have same data mapping pattern when call '%s'.\n",exp->symbol()->identifier(),dummy->identifier(),proc_ref->getHeader()->identifier());
				  	dummy_need_mapping = 0;
			      }
			    }
			}

			if (dummy_need_mapping != 0) 
			{
			  int detect_same_ok = check_reduced_dim_collapsed(exp);

			  if (opt_strategy & RUNTIMECOMM && detect_same_ok)
				dummy_need_mapping = 2;
			}

		    } // if (Dist_Def)

		      // further checking
		      // need to compare SHADOW & OUT_OF_CORE 
		    if (exp->variant() == VAR_REF || exp->variant() == ARRAY_REF) {
		        int same_other_directive = 1;
		    	SgSymbol *actusym = exp->symbol();
			      // compare out_of_core
		    	if (is_out_of_core(actusym) && !is_out_of_core(dummy) ||
			    !is_out_of_core(actusym) && is_out_of_core(dummy) )
			{
if (!silentFlag)
printf("Actual variable: %s and Dummy %s don't have same OUT_OF_CORE.\n",actusym->identifier(),dummy->identifier());
				same_other_directive = 0;
			}
			  // compare shadow
			  // inheirt dummy neednot to do so
			if (!is_inherit(dummy))
			if (same_other_directive == 1) 
			{
			      Shadow_def *sdef1 = get_defined_shadow(actusym);
			      Shadow_def *sdef2 = get_defined_shadow(dummy);
			      if (sdef1 && sdef2) {
				if (!compare_shadow(sdef1, sdef2)) 
					same_other_directive = 0;
			      }
		              else if (!sdef1 && !sdef2) ; 
			      else same_other_directive = 0;
			      if (same_other_directive == 0)
			      if (!silentFlag)
printf("Actual variable: %s and Dummy %s don't have same SHADOW.\n",actusym->identifier(),dummy->identifier());
			}
			  // need remap anyway
			if (same_other_directive == 0 )
			  dummy_need_mapping = 1;
		    }

		    if (dummy_need_mapping != 0)
		    {
/*
			  // IN_BIT need not remap after call
			if (dummy->attributes() & IN_BIT)
				*attr = -*attr;
*/
		     	exp->addAttribute(DUMMY_MAPPING,(void *)dummy_need_mapping,sizeof(int));
		    }

		    // is allowed to be called in procedure. ( consider reduced dimension )
		    // inter-procedure analysis
		    if (opt_strategy & OTHER_OPTIMIZE)
		    if (isSgArrayType(dummy->type()))
	            {
		       SgFunctionSymb *realHeader = proc_ref->getRealHeader();
		       SgSymbol *realDummy = get_real_dummy(realHeader, dummy);
		       if(realDummy)
		       {
		         if (isSgVarRefExp(exp) || dummy_need_mapping == 1)
		         {
if (!silentFlag)
printf("add ALL_COLLAPSED to dummy:%s\n", realDummy->identifier());
			  if (realDummy->numberOfAttributes(ALL_COLLAPSED) ==0 )
				realDummy->addAttribute(ALL_COLLAPSED,NULL,0);
		         }
		         else 
		         {
			  if (realDummy->numberOfAttributes(ALL_COLLAPSED) !=0 )
				realDummy->deleteAllAttribute(ALL_COLLAPSED);
                         }
		       }
                    }
		}	// end of for
	}  // end of for
}

void remap_parameter_with_core_shadow(SgStatement *s, SgExpression *actuDummy, SgSymbol *dadsrc, SgSymbol *dummy, SgSymbol *dadsec)
{
	SgSymbol *actusym = actuDummy->symbol();
	SgArrayType *arraysymbtype = isSgArrayType(actuDummy->type());
	ASSERT(arraysymbtype);
	int rank = arraysymbtype->dimension();
	Shadow_def *sdef = get_defined_shadow(dummy);

	  // new array data
	SgExpression *grpExp = make_group(s, dadsrc);
	SgStatement *new_data = make_new_array_data(s, dadsec, arraysymbtype,-1,rank,grpExp);
	s->insertStmtBefore(*new_data, *s->controlParent());

          // set_array_align
	for (int i=0; i<rank; i++)
	{
	SgExpression *start = get_lbound_from_arrayexp(actuDummy, i);
	SgExpression *end = get_ubound_from_arrayexp(actuDummy, i);
	SgExpression *a = new SgValueExp(1);
	SgExpression *b = new SgValueExp(0);
	int gl = sdef ? sdef->getLower(i):0;
	int gu = sdef ? sdef->getUpper(i):0;
	SgExpression *gtl = new SgValueExp(gl);
	SgExpression *gtu = new SgValueExp(gu);
	SgExpression *rng = make_range(s, dadsrc, new SgValueExp(i+1));
	SgStatement *align_stmt = make_set_array_align(s, dadsec, i+1, start, end, a ,b, gtl, gtu, rng);
	s->insertStmtBefore(*align_stmt, *s->controlParent());
	}

    	if ( is_out_of_core(dummy) )
        {
      	     SgStatement * ts = make_oc_init(s, dadsec, IN_CORE_SIZE);
	     s->insertStmtBefore(*ts, *s->controlParent());
	}
}

void add_dummy_mapping(SgStatement *s)
{
	SgStatement *scope = s->getScopeForDeclare();

	for (int i=0; i < s->numberOfAttributes(Proccall_Ref);i++)
	{
		proc_reference *proc_ref = (proc_reference *)s->attributeValue(i,Proccall_Ref);

		SgStatement *interface_hedr = proc_ref->getHeader()->function_hedr();

		int counter=0;
		for (int j=0; j< proc_ref->get_number() ; j ++)
		{
		   SgSymbol *dummy = proc_ref->get_dummy(j);
		   SgExpression *exp = proc_ref->get_actual(j);
		   if ( exp->variant() == ARRAY_REF ) counter++;
		   if (exp->numberOfAttributes(DUMMY_MAPPING) != 0)
		   {
			int dummy_mapping = (int)exp->attributeValue(0,DUMMY_MAPPING);

			SgSymbol *dadsrc,*dadsec;
			if (exp->variant() == VAR_REF)
  				dadsrc = s->getSymbolInCurrentScope(mixstring, "dad", exp->symbol()->identifier());
			else
  				dadsrc = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
  			dadsec = s->getSymbolInCurrentScope(mixstring, "tdd1", proc_ref->getHeader()->identifier(),counter,j);

			Distribute_def * ddef = NULL;
			SgSymbol * procname = NULL;
			Processor_def *pdef = NULL;
			if (dummy->numberOfAttributes(Dist_Def) == 0 ) 
			{
			  ASSERT(dummy->numberOfAttributes(Inherit_Def)>0 && dummy_mapping == 1);
printf("Warning:Inherit dummy '%s' with different Out-of-core or shadow.!\n", dummy->identifier());
			}
			else {
			  ddef = (Distribute_def *)dummy->attributeValue(0,Dist_Def);
			  procname = ddef->gettosymbol();
        		  pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
			}
			SgType *type = exp->type();


			  //  RUNTIME DATA MAPPING CHECKING
			  //    pcrc_same_data_mapping()
			  //  dummy_mapping == 2 
			if ( dummy_mapping == 2)
			// begin 
			{
if (!silentFlag)
printf("add_dummy_mapping:add runtime data mapping check of '%s' before call '%s'.\n",dummy->identifier(),proc_ref->getHeader()->identifier());

			SgStatement *if_stmt1,*if_stmt2,*last1,*last2;
			SgSymbol * need_comm = s->getSymbolInCurrentScope(mixstring, "comm", NULL, counter,j);
			SgExpression *needExp = new SgVarRefExp(need_comm);
			SgExpression *conExp = &( SgEqOp(*needExp,*new SgValueExp(0)));

			// create dummy dad to compare with actual arg dad
			// exp->symbol() must be an array
	    		SgSymbol * procgrd = pdef->getgrp();
			SgExpression *procExp = new SgVarRefExp(procgrd);
  			int storage = getArrayStorage(dummy);
			SgStatement *datastmt=make_new_array_data(s,dadsec,type,storage,-1,procExp);
			s->insertStmtBefore(*datastmt,*s->controlParent());
			distribute_array(scope,dadsec,exp,ddef,pdef,s,BEFORE);

			// check if same data mapping
			SgStatement *check_stmt=make_same_data_mapping(s,need_comm,dadsrc,dadsec);
			s->insertStmtBefore(*check_stmt,*s->controlParent());

			// true branch ; the last statement is remap() 
			SgStatement *remapstmt=make_remap(s,dadsec,dadsrc);
			last1 = remapstmt;

			// else branch ; first delete dadsec
			SgStatement *deletestmt=make_delete_array(s,dadsec);
			last2 = deletestmt;

			// if (comm .eq. 1) then ; else ;
  			if_stmt1 = new SgIfStmt(*conExp, *last1,*last2);
			s->insertStmtBefore(*if_stmt1,*s->controlParent());

			// in else branch, assign dadsec to dadsrc
			SgStatement *assignstmt = new SgAssignStmt(*new SgVarRefExp(dadsec), *new SgVarRefExp(dadsrc));
			last2->insertStmtAfter(*assignstmt,*last2->controlParent());

			SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", proc_ref->getHeader()->identifier(), counter,j);
#ifdef MEM_ALLOC
    			SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", NULL, counter,j);
			SgStatement *allocstmt=make_stack_alloc(s,type,base,dadsec,mHandle);
#else
			SgStatement *allocstmt = make_alloc(s,type,base,dadsec);
#endif
			last1->insertStmtBefore(*allocstmt,*last1->controlParent());
			SgStatement *resetstmt = make_reset_array_base(s,dadsec,type,base,mHandle);
			last1->insertStmtBefore(*resetstmt,*last1->controlParent());

			// remap after call
			remapstmt=make_remap(s,dadsrc,dadsec);
  			if_stmt2 = new SgIfStmt(*conExp, *remapstmt);
			last1 = remapstmt;
			s->insertStmtAfter(*if_stmt2,*s->controlParent());
#ifdef MEM_ALLOC
			base = s->getSymbolInCurrentScope(mixstring, "mh", NULL, counter,j);
#endif
			SgStatement *freestmt=make_free(s,type,base);
			last1->insertStmtAfter(*freestmt,*last1->controlParent());

			// delete array after if stmt
			deletestmt=make_delete_array(s,dadsec);
			last1->insertStmtAfter(*deletestmt,*last1->controlParent());

			// get index from  pcrc_integer_index()
			SgSymbol * newindex = s->getSymbolInCurrentScope(mixstring, "sdx", NULL, counter,j);
			SgStatement *indexstmt=make_index(s,newindex,dadsec,type);
			s->insertStmtBefore(*indexstmt,*s->controlParent());

			// add attributes to exp
//			exp->addAttribute(DUMMY_ADDR,((SgCallStmt*)resetstmt)->arg(1),sizeof(SgExpression *) );
#ifdef MEM_ALLOC
			SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(type),"mb", NULL); 
#else
			SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(type), NULL); 
#endif
			SgArrayRefExp *stackexp = new SgArrayRefExp(*stack);
			SgExpression *idx = new SgVarRefExp(*newindex);
			stackexp->addSubscript(*idx);
			exp->addAttribute(DUMMY_ADDR,stackexp,sizeof(SgExpression *) );
			exp->addAttribute(DUMMY_NAME,dadsec,sizeof(dadsec));
			}
			//end
			  //  REMAP DUMMY TO TEMPORARY DAD
			else if (dummy_mapping	== 1) // no optimize
			// begin 
			{
if (!silentFlag)
printf("add_dummy_mapping:distributed '%s' before call '%s'.\n",dummy->identifier(),proc_ref->getHeader()->identifier());
			// exp->symbol() must be an array
			////// here dummy must have distribute attributes
			if (ddef != NULL) {
			  SgStatement *datastmt=make_new_array_data(s,dadsec,type);
			  s->insertStmtBefore(*datastmt,*s->controlParent());
			  distribute_array(scope,dadsec,exp,ddef,pdef,s,BEFORE);
			}
			else
			{
			  // so, dummy may have inherit, but have different shadow or out_of_core attributes
			  remap_parameter_with_core_shadow(s, exp, dadsrc, dummy, dadsec);
			}
			SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", "m", counter,j);
#ifdef MEM_ALLOC
    			SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", NULL, counter,j);
			SgStatement *allocstmt=make_stack_alloc(s,type,base,dadsec,mHandle);
#else
			SgStatement *allocstmt = make_alloc(s,type,base,dadsec);
#endif
			s->insertStmtBefore(*allocstmt,*s->controlParent());
			SgStatement *resetstmt = make_reset_array_base(s,dadsec,type,base,mHandle);
			s->insertStmtBefore(*resetstmt,*s->controlParent());

			SgStatement *remapstmt=make_remap(s,dadsec,dadsrc);
			s->insertStmtBefore(*remapstmt,*s->controlParent());

#ifdef MEM_ALLOC
			base = s->getSymbolInCurrentScope(mixstring, "mh", NULL, counter,j);
#endif

			SgStatement *freestmt=make_free(s,type,base);
			s->insertStmtAfter(*freestmt,*s->controlParent());

			SgStatement *deletestmt=make_delete_array(s,dadsec);
			s->insertStmtAfter(*deletestmt,*s->controlParent());

			remapstmt=make_remap(s,dadsrc,dadsec);
			s->insertStmtAfter(*remapstmt,*s->controlParent());
			exp->addAttribute(DUMMY_ADDR,((SgCallStmt*)resetstmt)->arg(1),sizeof(SgExpression *) );
			exp->addAttribute(DUMMY_NAME,dadsec,sizeof(dadsec));
			}
			//end
		    } // else
		}	// end of for
	}  // end of for
}


